import bisect
import datetime
import time

from pydgn.training.event.handler import EventHandler


# Decorator class. Ideas taken from "Fluent Python" book
class Profiler:
    r"""
    A decorator class that is applied to a :class:`~pydgn.training.event.handler.EventHandler` object implementing a
    set of callback functions. For each callback, the Profiler stores the average and total running time across epochs.
    When the experiment terminates (either correctly or abruptly) the Profiler can produce a report to be stored in the
    experiment's log file.

    The Profiler is used as a singleton, and it produces wrappers that update its own state.

    Args:
        threshold (float): used to filter out callback functions that consume a negligible amount of time from the report

    Usage:
        Istantiate a profiler, and then register an event_handler with the syntax profiler(event_handler), which returns
        another object implementing the :class:`~pydgn.training.event.handler.EventHandler` interface
    """
    def __init__(self, threshold: float):
        # we filter out computation that takes a negligible amount of time from the report (< threshold)
        self.threshold = threshold
        self.callback_elapsed = {}
        self.callback_calls = {}

    def __call__(self, event_handler: EventHandler) -> object:
        r"""
        Wraps a :class:`~pydgn.training.event.handler.EventHandler` object, so that whenever one of its callbacks is
        triggered the Profiler can compute and store statistics about the time required to execute it.

        Args:
            event_handler (:class:`~pydgn.training.event.handler.EventHandler`): the object implementing a subset of
                          the callbacks defined in the :class:`~pydgn.training.event.handler.EventHandler` interface

        Returns:
            an object that transparently updates the profiler's state whenever a callback is triggered
        """
        self.callback_elapsed[event_handler.__class__.__name__] = {}
        self.callback_calls[event_handler.__class__.__name__] = {}

        def clock(func):
            def clocked(*args, **kwargs):
                class_name = event_handler.__class__.__name__
                callback_name = func.__name__
                t0 = time.time()
                # print(f'Callback {cls.__class__.__name__}: calling {func.__name__} with args {args} and kwargs {kwargs}')
                result = func(*args, **kwargs)
                elapsed = time.time() - t0

                if callback_name in self.callback_elapsed[class_name]:
                    self.callback_elapsed[class_name][callback_name] += elapsed
                    self.callback_calls[class_name][callback_name] += 1
                else:
                    self.callback_elapsed[class_name][callback_name] = elapsed
                    self.callback_calls[class_name][callback_name] = 1

                return result

            return clocked

        # Taken from: https://stackoverflow.com/questions/3155436/getattr-for-static-class-variables-in-python
        # We need to do this to invoke __getattr__ on a class
        # https://stackoverflow.com/questions/100003/what-are-metaclasses-in-python
        # TLDR We want the class ClockedCallback (which is an object itself) to have a method __getattr__
        # if we specify __getattr__ inside ClockedCallback, rather than in its metaclass, the class ClockedCallback will
        # not use the overridden __getattr__
        class getattribute(type):
            def __getattr__(self, name):
                return clock(getattr(event_handler, name)) #  needed to implement callback calls with generic names

        # Relies on closures
        class ClockedCallback(metaclass=getattribute):
            pass

        return ClockedCallback

    def report(self) -> str:
        r"""
        Builds a report string containing the statistics of the experiment accumulated so far.

        Returns:
            a string containing the report
        """
        total_time_experiment = 0.
        profile_str = f'{"*" * 25} Profiler {"*" * 25} \n \n'
        profile_str += f'Threshold: {self.threshold} \n \n'

        for class_name, v in self.callback_elapsed.items():

            sorted_avg_elapsed = []

            for callback_name, v1 in v.items():
                n = self.callback_calls[class_name][callback_name]

                # Depending on when a KeyboardInterrupt is triggered
                if n == 0:
                    continue

                avg_elapsed = v1 / n

                if avg_elapsed > self.threshold:
                    bisect.insort(sorted_avg_elapsed, (avg_elapsed, v1, callback_name))

            # Release resources
            v.clear()

            profile_str += f'{class_name} \n \n'

            for (avg_elapsed, total_elapsed, callback_name) in reversed(sorted_avg_elapsed):
                total_time_experiment += total_elapsed
                profile_str += f'\t {callback_name} --> Avg: {avg_elapsed} s, Total: {str(datetime.timedelta(seconds=total_elapsed))} \n'
            profile_str += '\n'

        profile_str += f'Total time of the experiment: {str(datetime.timedelta(seconds=total_time_experiment))} \n \n'
        profile_str += f'{"*" * 60}'
        return profile_str



from typing import List, Optional, Tuple

import torch
from torch import nn
from torch_geometric.data import Batch
from torch_geometric.nn import global_add_pool

from pydgn.model.interface import ModelInterface


class ToyMLP(ModelInterface):
    """
    A toy MLP model used to test the library. Technically, a DGN that ignores the adjacency matrix.
    """
    def __init__(self, dim_node_features, dim_edge_features, dim_target, readout_class, config):
        super().__init__(dim_node_features, dim_edge_features, dim_target, readout_class, config)

        dim_embedding = config['dim_embedding']
        self.W = nn.Linear(dim_node_features, dim_target, bias=True)


    def forward(self, data: Batch) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        x, batch = data.x , data.batch

        hg = global_add_pool(x, batch)
        return self.W(hg), x


import torch.nn as nn

from pydgn.model.interface import ReadoutInterface


class LinearNodeReadout(ReadoutInterface):
    """
    Class that implements a simple readout mapping for node prediction
    """
    def __init__(self, dim_node_features, dim_edge_features, dim_target, config):
        super().__init__(dim_node_features, dim_edge_features, dim_target, config)
        self.W = nn.Linear(dim_node_features, dim_target, bias=True)

    def forward(self, node_embeddings, batch, **kwargs):
        return self.W(node_embeddings), node_embeddings



import copy

from pydgn.experiment.util import s2c
from pydgn.model.interface import ModelInterface
from pydgn.training.event.handler import EventHandler


class Optimizer(EventHandler):
    """
    Optimizer is the main event handler for optimizers.
    Just pass a PyTorch optimizer together with its arguments in the configuration file.

    Args:
        model (:class:`~pydgn.model.interface.ModelInterface`): the model that has to be trained
        optimizer_class_name (str): dotted path to the optimizer class to use
        accumulate_gradients (bool): if ``True``, accumulate mini-batch gradients to perform a batch gradient update without loading the entire batch in memory
        kwargs (dict): additional parameters for the specific optimizer
    """
    def __init__(self,
                 model: ModelInterface,
                 optimizer_class_name: str,
                 accumulate_gradients: bool=False,
                 **kwargs: dict):
        super().__init__()
        self.optimizer = s2c(optimizer_class_name)(model.parameters(), **kwargs)
        self.accumulate_gradients = accumulate_gradients

    def load_state_dict(self, state_dict):
        self.optimizer.load_state_dict(state_dict)

    def on_fit_start(self, state):
        if state.optimizer_state is not None:
            self.optimizer.load_state_dict(state.optimizer_state)

    def on_training_epoch_start(self, state):
        if self.accumulate_gradients:
            self.optimizer.zero_grad()

    def on_training_batch_start(self, state):
        if not self.accumulate_gradients:
            self.optimizer.zero_grad()

    def on_training_batch_end(self, state):
        if not self.accumulate_gradients:
            self.optimizer.step()

    def on_training_epoch_end(self, state):
        if self.accumulate_gradients:
            self.optimizer.step()

    def on_epoch_end(self, state):
        state.update(optimizer_state=copy.deepcopy(self.optimizer.state_dict()))



import os
OMP_NUM_THREADS = 'OMP_NUM_THREADS'
# TODO do we still need this?
os.environ[OMP_NUM_THREADS] = "1"  # This is CRUCIAL to avoid bottlenecks when running experiments in parallel. DO NOT REMOVE IT
import yaml
import argparse

import ray
import torch

# TODO do we still need this?
# Needed to avoid thread spawning, conflicts with multi-processing. You may set a number > 1 but take into account the number of processes on the machine
torch.set_num_threads(1)

from pydgn.static import *
from pydgn.experiment.util import s2c
from pydgn.data.splitter import Splitter
from pydgn.evaluation.grid import Grid
from pydgn.evaluation.random_search import RandomSearch
from pydgn.evaluation.util import set_gpus


def evaluation(options):
    kwargs = vars(options)
    debug = kwargs[DEBUG]
    configs_dict = yaml.load(open(kwargs[CONFIG_FILE], "r"), Loader=yaml.FullLoader)

    assert GRID_SEARCH in configs_dict or RANDOM_SEARCH in configs_dict
    search_class = (Grid if GRID_SEARCH in configs_dict else RandomSearch)
    search = search_class(configs_dict)

    # Set the random seed
    seed = search.seed if search.seed is not None else 42
    print(f'Base seed set to {seed}.')
    experiment = search.experiment
    experiment_class = s2c(experiment)
    use_cuda = CUDA in search.device
    exp_path = os.path.join(configs_dict[RESULT_FOLDER], f"{search.exp_name}")

    # Hardware initialization
    max_cpus = configs_dict[MAX_CPUS]

    if not use_cuda:
        max_gpus = 0
        gpus_per_task = 0
    else:
        # Ensure a generic "cuda" device is set when using more than 1 GPU
        # We will choose the GPU with least ratio of memory usage
        search.device = CUDA
        max_gpus = configs_dict[MAX_GPUS]
        # Choose which GPUs to use
        set_gpus(max_gpus)
        gpus_per_task = configs_dict[GPUS_PER_TASK]

    os.environ[PYDGN_RAY_NUM_GPUS_PER_TASK] = str(int(gpus_per_task))

    # You can make PyDGN work on a cluster of machines!
    if os.environ.get('ip_head') is not None:
        assert os.environ.get('redis_password') is not None
        ray.init(address=os.environ.get('ip_head'), _redis_password=os.environ.get('redis_password'))
        print("Connected to Ray cluster.")
        print(f"Available nodes: {ray.nodes()}")
    # Or you can work on your single server
    else:
        ray.init(num_cpus=max_cpus, num_gpus=max_gpus)
        print(f"Started local ray instance.")

    data_splits_file = configs_dict[DATA_SPLITS_FILE]

    splitter = Splitter.load(data_splits_file)
    inner_folds, outer_folds = splitter.n_inner_folds, splitter.n_outer_folds
    print(f'Data splits loaded, outer folds are {outer_folds} and inner folds are {inner_folds}')

    # WARNING: leave the import here, it reads env variables set before
    from pydgn.evaluation.evaluator import RiskAssesser
    risk_assesser = RiskAssesser(outer_folds,
                                 inner_folds,
                                 experiment_class,
                                 exp_path,
                                 data_splits_file,
                                 search,
                                 final_training_runs=configs_dict[FINAL_TRAINING_RUNS],
                                 higher_is_better=search.higher_results_are_better,
                                 gpus_per_task=gpus_per_task,
                                 base_seed=seed)

    risk_assesser.risk_assessment(debug=debug)


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(CONFIG_FILE_CLI_ARGUMENT, dest=CONFIG_FILE)
    parser.add_argument(DEBUG_CLI_ARGUMENT, action="store_true", dest=DEBUG, default=False)
    return parser.parse_args()


def main():
    options = get_args()
    try:
        evaluation(options)
    except Exception as e:
        raise e



from typing import List

import torch
from torch.nn import Module, CrossEntropyLoss, MSELoss

from pydgn.experiment.util import s2c
from pydgn.static import BATCH_LOSS_EXTRA, ARGS, CLASS_NAME
from pydgn.training.event.handler import EventHandler
from pydgn.training.event.state import State


class Metric(Module, EventHandler):
    r"""
    Metric is the main event handler for all metrics. Other metrics can easily subclass by implementing the :func:`forward`
    method, though sometimes more complex implementations are required.

    Args:
        use_as_loss (bool): whether this metric should act as a loss (i.e., it should act when :func:`on_backward` is called). **Used by PyDGN, no need to care about this.**
        reduction (str): the type of reduction to apply across samples of the mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        use_nodes_batch_size (bool): whether or not to use the # of nodes in the batch, rather than the number of graphs, to compute
        the metric's aggregated value for the entire epoch.
    """
    def __init__(self, use_as_loss: bool=False, reduction: str='mean', use_nodes_batch_size: bool=False):
        super().__init__()
        self.batch_metrics = None
        self.num_samples = None
        self.use_as_loss = use_as_loss
        self.reduction = reduction
        self.use_nodes_batch_size = use_nodes_batch_size

    @property
    def name(self) -> str:
        raise NotImplementedError('You should subclass Metric and implement this method!')

    def get_main_metric_name(self) -> str:
        """
        Return the metric's main name. Useful when a metric is the combination of many.

        Returns:
            the metric's main name
        """
        return self.name

    def _handle_reduction(self, state: State):
        #
        # Helper function. it provides the exact value to use when updating the metric's counters
        #
        if self.reduction == 'mean':
            # Used to recover the "sum" version of the metric
            return state.batch_num_targets if not self.use_nodes_batch_size else state.batch_num_nodes

        elif self.reduction == 'sum':
            return 1
        else:
            raise NotImplementedError('The only reductions allowed are sum and mean')


    def on_training_epoch_start(self, state: State):
        self.batch_metrics = []
        self.num_samples = 0

    def on_training_batch_end(self, state: State):
        if self.use_as_loss:
            batch_metric = state.batch_loss
        else:
            batch_metric = state.batch_score

        self.batch_metrics.append(batch_metric[self.name].item() * self._handle_reduction(state))
        self.num_samples += state.batch_num_targets if not self.use_nodes_batch_size else state.batch_num_nodes

    def on_training_epoch_end(self, state: State):
        if self.use_as_loss:
            state.update(epoch_loss={self.name: torch.tensor(self.batch_metrics).sum() / self.num_samples})
        else:
            state.update(epoch_score={self.name: torch.tensor(self.batch_metrics).sum() / self.num_samples})

        self.batch_metrics = None
        self.num_samples = None

    def on_eval_epoch_start(self, state: State):
        self.batch_metrics = []
        self.num_samples = 0

    def on_eval_epoch_end(self, state: State):
        if self.use_as_loss:
            state.update(epoch_loss={self.name: torch.tensor(self.batch_metrics).sum() / self.num_samples})
        else:
            state.update(epoch_score={self.name: torch.tensor(self.batch_metrics).sum() / self.num_samples})

        self.batch_metrics = None
        self.num_samples = None

    def on_eval_batch_end(self, state: State):
        if self.use_as_loss:
            batch_metric = state.batch_loss
        else:
            batch_metric = state.batch_score

        self.batch_metrics.append(batch_metric[self.name].item() * self._handle_reduction(state))
        self.num_samples += state.batch_num_targets if not self.use_nodes_batch_size else state.batch_num_nodes

    def on_compute_metrics(self, state: State):
        outputs, targets = state.batch_outputs, state.batch_targets

        if self.use_as_loss:
            loss_output = self.forward(targets, *outputs)

            if isinstance(loss_output, tuple):
                # Allow loss to produce intermediate results that speed up
                # Score computation. This loss callback MUST occur before the score one.
                loss, extra = loss_output
                state.update(batch_loss_extra={self.name: extra})
            else:
                loss = loss_output

            state.update(batch_loss={self.name: loss})

        else:
            score = self.forward(targets,
                                 *outputs,
                                 batch_loss_extra=getattr(state, BATCH_LOSS_EXTRA, None))

            # Handle base case: forward returns a score
            # This is used to work with implementations that return multiple scores, such as MultiScore
            if type(score) is not dict:
                score = {self.name: score}

            # Score is a dictionary with key-value pairs
            # we need to detach each score from the computational graph
            score = {k: v.detach().cpu() for k, v in score.items()}
            state.update(batch_score=score)

    def on_backward(self, state: State):
        if self.use_as_loss:
            try:
                state.batch_loss[self.name].backward()
            except Exception as e:
                # Here we catch potential multiprocessing related issues
                # see https://github.com/pytorch/pytorch/wiki/Autograd-and-Fork
                print(e)

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        r"""
        Computes the metric value. Optionally, and only for scores used as losses, some extra information can be also returned.

        Args:
            targets (:class:`torch.Tensor`): ground truth
            outputs (List[:class:`torch.Tensor`]): outputs of the model
            batch_loss_extra (dict): dictionary of information computed by metrics used as losses

        Returns:
            A dictionary containing associations metric_name - value
        """
        raise NotImplementedError('To be implemented by a subclass')


class MultiScore(Metric):
    r"""
    This class is used to keep track of multiple additional metrics used as scores, rather than losses.

    Args:
        use_as_loss (bool): whether this metric should act as a loss (i.e., it should act when :func:`on_backward` is called). **Used by PyDGN, no need to care about this.**
        reduction (str): the type of reduction to apply across samples of the mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        use_nodes_batch_size (bool): whether or not to use the # of nodes in the batch, rather than the number of graphs, to compute
        the metric's aggregated value for the entire epoch.
    """
    def __init__(self, use_as_loss, reduction='mean', use_nodes_batch_size=False, main_scorer=None, **extra_scorers):
        assert not use_as_loss, "MultiScore cannot be used as loss"
        assert not main_scorer is None, "You have to provide a main scorer"
        super().__init__(use_as_loss, reduction, use_nodes_batch_size)
        self.scorers = [self._istantiate_scorer(main_scorer)] + [self._istantiate_scorer(score) for score in
                                                                 extra_scorers.values()]

    @property
    def name(self) -> str:
        return "Multi Score"

    def _istantiate_scorer(self, scorer):
        if isinstance(scorer, dict):
            args = scorer[ARGS]
            return s2c(scorer[CLASS_NAME])(use_as_loss=False, **args)
        else:
            return s2c(scorer)(use_as_loss=False)

    def get_main_metric_name(self):
        return self.scorers[0].get_main_metric_name()

    def on_training_epoch_start(self, state: State):
        self.batch_metrics = {s.name: [] for s in self.scorers}
        for scorer in self.scorers:
            scorer.on_training_epoch_start(state)

    def on_training_batch_end(self, state: State):
        for scorer in self.scorers:
            scorer.on_training_batch_end(state)

    def on_training_epoch_end(self, state: State):
        epoch_score = {}
        for scorer in self.scorers:
            # This will update the epoch_score variable in State
            scorer.on_training_epoch_end(state)
            epoch_score.update(state.epoch_score)
        state.update(epoch_score=epoch_score)

    def on_eval_epoch_start(self, state: State):
        for scorer in self.scorers:
            scorer.on_eval_epoch_start(state)

    def on_eval_batch_end(self, state: State):
        for scorer in self.scorers:
            scorer.on_eval_batch_end(state)

    def on_eval_epoch_end(self, state: State):
        epoch_score = {}
        for scorer in self.scorers:
            # This will update the epoch_score variable in State
            scorer.on_training_epoch_end(state)
            epoch_score.update(state.epoch_score)
        state.update(epoch_score=epoch_score)

    def on_compute_metrics(self, state: State):
        super().on_compute_metrics(state)

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        res = {}
        for scorer in self.scorers:
            # each scorer __call__ method returns a dict
            res.update({scorer.name: scorer.forward(targets, *outputs, batch_loss_extra=batch_loss_extra)})

        return res


class AdditiveLoss(Metric):
    r"""
    AdditiveLoss sums an arbitrary number of losses together.

    Args:
        use_as_loss (bool): whether this metric should act as a loss (i.e., it should act when :func:`on_backward` is called). **Used by PyDGN, no need to care about this.**
        reduction (str): the type of reduction to apply across samples of the mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        use_nodes_batch_size (bool): whether or not to use the # of nodes in the batch, rather than the number of graphs, to compute
        the metric's aggregated value for the entire epoch.
        losses (dict): dictionary of metrics to add together
    """
    def __init__(self, use_as_loss, reduction='mean', use_nodes_batch_size=False, **losses: dict):
        assert use_as_loss, "Additive loss can only be used as a loss"
        super().__init__(use_as_loss, reduction, use_nodes_batch_size)

        self.losses = [self._instantiate_loss(loss) for loss in losses.values()]
        self.use_nodes_batch_size = use_nodes_batch_size

    @property
    def name(self) -> str:
        return "Additive Loss"

    def _instantiate_loss(self, loss):
        if isinstance(loss, dict):
            args = loss[ARGS]
            return s2c(loss[CLASS_NAME])(use_as_loss=True, **args)
        else:
            return s2c(loss)(use_as_loss=True)

    def on_training_epoch_start(self, state: State):
        self.batch_metrics = {l.name: [] for l in [self] + self.losses}
        self.num_samples = 0

    def on_training_batch_end(self, state: State):
        for k, v in state.batch_loss.items():
            self.batch_metrics[k].append(v.item() * self._handle_reduction(state))
        self.num_samples += state.batch_num_targets if not self.use_nodes_batch_size else state.batch_num_nodes

    def on_training_epoch_end(self, state: State):
        state.update(epoch_loss={l.name: torch.tensor(self.batch_metrics[l.name]).sum() / self.num_samples
                                 for l in [self] + self.losses})
        self.batch_metrics = None
        self.num_samples = None

    def on_eval_epoch_start(self, state: State):
        self.batch_metrics = {l.name: [] for l in [self] + self.losses}
        self.num_samples = 0

    def on_eval_epoch_end(self, state: State):
        state.update(epoch_loss={l.name: torch.tensor(self.batch_metrics[l.name]).sum() / self.num_samples
                                 for l in [self] + self.losses})
        self.batch_metrics = None
        self.num_samples = None

    def on_eval_batch_end(self, state: State):
        for k, v in state.batch_loss.items():
            self.batch_metrics[k].append(v.item() * self._handle_reduction(state))
        self.num_samples += state.batch_num_targets if not self.use_nodes_batch_size else state.batch_num_nodes

    def on_compute_metrics(self, state: State):
        outputs, targets = state.batch_outputs, state.batch_targets
        loss = {}
        extra = {}
        loss_sum = 0.
        for l in self.losses:
            single_loss = l.forward(targets, *outputs)
            if isinstance(single_loss, tuple):
                # Allow loss to produce intermediate results that speed up
                # Score computation. Loss callback MUST occur before the score one.
                loss_output, loss_extra = single_loss
                extra[l.name] = loss_extra
                state.update(batch_loss_extra=extra)
            else:
                loss_output = single_loss
            loss[l.name] = loss_output
            loss_sum += loss_output

        loss[self.name] = loss_sum
        state.update(batch_loss=loss)

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        pass
    

class Classification(Metric):
    r"""
    Generic metric for classification tasks. Used to maximize code reuse for classical metrics.

    Args:
        use_as_loss (bool): whether this metric should act as a loss (i.e., it should act when :func:`on_backward` is called). **Used by PyDGN, no need to care about this.**
        reduction (str): the type of reduction to apply across samples of the mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        use_nodes_batch_size (bool): whether or not to use the # of nodes in the batch, rather than the number of graphs, to compute
        the metric's aggregated value for the entire epoch.
    """
    def __init__(self, use_as_loss=False, reduction='mean', use_nodes_batch_size=False):
        super().__init__(use_as_loss=use_as_loss, reduction=reduction, use_nodes_batch_size=use_nodes_batch_size)
        self.metric = None

    @property
    def name(self) -> str:
        return 'Classification Metric'

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        outputs = outputs[0]

        if len(targets.shape) == 2:
            targets = targets.squeeze(dim=1)

        metric = self.metric(outputs.squeeze(dim=1), targets)
        return metric


class Regression(Metric):
    r"""
    Generic metric for regression tasks. Used to maximize code reuse for classical metrics.

    Args:
        use_as_loss (bool): whether this metric should act as a loss (i.e., it should act when :func:`on_backward` is called). **Used by PyDGN, no need to care about this.**
        reduction (str): the type of reduction to apply across samples of the mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        use_nodes_batch_size (bool): whether or not to use the # of nodes in the batch, rather than the number of graphs, to compute
        the metric's aggregated value for the entire epoch.
    """
    def __init__(self, use_as_loss=False, reduction='mean', use_nodes_batch_size=False):
        super().__init__(use_as_loss=use_as_loss, reduction=reduction, use_nodes_batch_size=use_nodes_batch_size)
        self.metric = None

    @property
    def name(self) -> str:
        return 'Regression Metric'

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        outputs = outputs[0]

        metric = self.metric(outputs.squeeze(dim=1), targets.squeeze(dim=1))
        return metric


class MulticlassClassification(Classification):
    r"""
    Wrapper around :class:`torch.nn.CrossEntropyLoss`

    Args:
        use_as_loss (bool): whether this metric should act as a loss (i.e., it should act when :func:`on_backward` is called). **Used by PyDGN, no need to care about this.**.
        reduction (str): the type of reduction to apply across samples of the mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        use_nodes_batch_size (bool): whether or not to use the # of nodes in the batch, rather than the number of graphs, to compute
        the metric's aggregated value for the entire epoch.
    """
    def __init__(self, use_as_loss=False, reduction='mean', use_nodes_batch_size=False):
        super().__init__(use_as_loss=use_as_loss, reduction=reduction, use_nodes_batch_size=use_nodes_batch_size)
        self.metric = CrossEntropyLoss(reduction=reduction)

    @property
    def name(self) -> str:
        return 'Multiclass Classification'

class MeanSquareError(Regression):
    r"""
    Wrapper around :class:`torch.nn.MSELoss`

    Args:
        use_as_loss (bool): whether this metric should act as a loss (i.e., it should act when :func:`on_backward` is called). **Used by PyDGN, no need to care about this.**.
        reduction (str): the type of reduction to apply across samples of the mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        use_nodes_batch_size (bool): whether or not to use the # of nodes in the batch, rather than the number of graphs, to compute
        the metric's aggregated value for the entire epoch.
    """
    def __init__(self, use_as_loss=False, reduction='mean', use_nodes_batch_size=False):
        super().__init__(use_as_loss=use_as_loss, reduction=reduction, use_nodes_batch_size=use_nodes_batch_size)
        self.metric = MSELoss(reduction=reduction)

    @property
    def name(self) -> str:
        return 'Mean Square Error'


class DotProductLink(Metric):
    """
    Implements a dot product link prediction metric, as defined in https://arxiv.org/abs/1611.07308.
    """
    @property
    def name(self) -> str:
        return 'Dot Product Link Prediction'

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        node_embs = outputs[1]
        _, pos_edges, neg_edges = targets

        loss_edge_index = torch.cat((pos_edges, neg_edges), dim=1)
        loss_target = torch.cat((torch.ones(pos_edges.shape[1]),
                                 torch.zeros(neg_edges.shape[1])))

        # Taken from https://github.com/rusty1s/pytorch_geometric/blob/master/examples/link_pred.py
        x_j = torch.index_select(node_embs, 0, loss_edge_index[0])
        x_i = torch.index_select(node_embs, 0, loss_edge_index[1])
        link_logits = torch.einsum("ef,ef->e", x_i, x_j)
        loss = torch.nn.functional.binary_cross_entropy_with_logits(link_logits, loss_target.to(link_logits.device))
        return loss


class MulticlassAccuracy(Metric):
    """
    Implements multiclass classification accuracy.
    """

    @property
    def name(self) -> str:
        return 'Multiclass Accuracy'

    @staticmethod
    def _get_correct(output):
        return torch.argmax(output, dim=1)

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        pred = outputs[0]
        correct = self._get_correct(pred)
        return 100. * (correct == targets).sum().float() / targets.size(0)


class ToyMetric(Metric):
    """
    Implements a toy metric.

    Args:
        use_as_loss (bool): whether this metric should act as a loss (i.e., it should act when :func:`on_backward` is called). **Used by PyDGN, no need to care about this.**.
        reduction (str): the type of reduction to apply across samples of the mini-batch. Supports ``mean`` and ``sum``. Default is ``mean``.
        use_nodes_batch_size (bool): whether or not to use the # of nodes in the batch, rather than the number of graphs, to compute
        the metric's aggregated value for the entire epoch.
    """

    def __init__(self, use_as_loss=False, reduction='mean', use_nodes_batch_size=False):
        super().__init__(use_as_loss=use_as_loss, reduction=reduction, use_nodes_batch_size=use_nodes_batch_size)


    @property
    def name(self) -> str:
        return 'Toy Metric'

    @staticmethod
    def _get_correct(output):
        return torch.argmax(output, dim=1)

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict=None) -> dict:
        return 100. * torch.ones(1)


class ToyUnsupervisedMetric(Metric):

    @property
    def name(self) -> str:
        return 'Toy Unsupervised Metric'

    def forward(self,
                targets: torch.Tensor,
                *outputs: List[torch.Tensor],
                batch_loss_extra: dict = None) -> dict:
        return (outputs[0]*0.).mean()


import argparse

import yaml

from pydgn.data.util import preprocess_data
from pydgn.static import CONFIG_FILE_CLI_ARGUMENT, CONFIG_FILE


def get_args_dict():
    parser = argparse.ArgumentParser()
    parser.add_argument(CONFIG_FILE_CLI_ARGUMENT, dest=CONFIG_FILE, help='config file to parse the data')
    return vars(parser.parse_args())


def main():
    args = get_args_dict()
    options = yaml.load(open(args[CONFIG_FILE], "r"), Loader=yaml.FullLoader)
    preprocess_data(options)



import os
import shutil
from typing import List, Union, Tuple, Optional, Callable

import torch
import torch_geometric
from ogb.graphproppred import PygGraphPropPredDataset
from ogb.utils.url import decide_download, download_url, extract_zip
from torch_geometric.data import InMemoryDataset, Data
from torch_geometric.datasets import TUDataset, Planetoid


class ZipDataset(torch.utils.data.Dataset):
    r"""
    This Dataset takes `n` datasets and "zips" them. When asked for the `i`-th element, it returns the `i`-th element of
    all `n` datasets.

    Args:
        datasets (List[torch.utils.data.Dataset]): An iterable with PyTorch Datasets

    Precondition:
        The length of all datasets must be the same

    """
    def __init__(self, datasets: List[torch.utils.data.Dataset]):
        self.datasets = datasets
        assert len(set(len(d) for d in self.datasets)) == 1

    def __getitem__(self, index: int) -> List[object]:
        r"""
        Returns the `i`-th element of all datasets in a list

        Args:
            index (int): the index of the data point to retrieve from each dataset

        Returns:
            A list containing one element for each dataset in the ZipDataset
        """
        return [d[index] for d in self.datasets]

    def __len__(self) -> int:
        r"""
        Returns the length of the datasets (all of them have the same length)

        Returns:
            The unique number of samples of all datasets
        """
        return len(self.datasets[0])


class ConcatFromListDataset(InMemoryDataset):
    r"""
    Create a dataset from a list of :class:`torch_geometric.data.Data` objects. Inherits from
    :class:`torch_geometric.data.InMemoryDataset`

    Args:
        data_list (list): List of graphs.
    """
    def __init__(self, data_list: List[Data]):
        super(ConcatFromListDataset, self).__init__("")
        self.data, self.slices = self.collate(data_list)

    def download(self):
        pass

    def process(self):
        pass

    @property
    def processed_file_names(self) -> Union[str, List[str], Tuple]:
        return []

    @property
    def raw_file_names(self) -> Union[str, List[str], Tuple]:
        return []



class DatasetInterface(torch_geometric.data.dataset.Dataset):
    r"""
    Class that defines a number of properties essential to all datasets implementations inside PyDGN. These properties
    are used by the training engine and forwarded to the model to be trained. For some datasets, e.g.,
    :class:`torch_geometric.datasets.TUDataset`, implementing this interface is straightforward.

    Args:
        root (str): root folder where to store the dataset
        name (str): name of the dataset
        transform (Optional[Callable]): transformations to apply to each ``Data`` object at run time
        pre_transform (Optional[Callable]): transformations to apply to each ``Data`` object at dataset creation time
        pre_filter (Optional[Callable]): sample filtering to apply to each ``Data`` object at dataset creation time
    """
    def __init__(self,
                 root: str,
                 name: str,
                 transform: Optional[Callable]=None,
                 pre_transform: Optional[Callable]=None,
                 pre_filter: Optional[Callable]=None):

        self.name = name
        super().__init__(root, transform, pre_transform, pre_filter)

    @property
    def raw_file_names(self) -> Union[str, List[str], Tuple]:
        raise NotImplementedError("You should subclass DatasetInterface and implement this method")

    @property
    def processed_file_names(self) -> Union[str, List[str], Tuple]:
        raise NotImplementedError("You should subclass DatasetInterface and implement this method")

    def download(self):
        raise NotImplementedError("You should subclass DatasetInterface and implement this method")

    def process(self):
        raise NotImplementedError("You should subclass DatasetInterface and implement this method")

    def len(self) -> int:
        raise NotImplementedError("You should subclass DatasetInterface and implement this method")

    def get(self, idx: int) -> Data:
        raise NotImplementedError("You should subclass DatasetInterface and implement this method")

    @property
    def dim_node_features(self):
        r"""
        Specifies the number of node features (after pre-processing, but in the end it depends on the model that is
        implemented).
        """
        raise NotImplementedError("You should subclass DatasetInterface and implement this method")

    @property
    def dim_edge_features(self):
        r"""
        Specifies the number of edge features (after pre-processing, but in the end it depends on the model that is
        implemented).
        """
        raise NotImplementedError("You should subclass DatasetInterface and implement this method")

    @property
    def dim_target(self):
        r"""
        Specifies the dimension of each target vector.
        """
        raise NotImplementedError("You should subclass DatasetInterface and implement this method")

    def __len__(self) -> int:
        raise NotImplementedError("You should subclass DatasetInterface and implement this method")


class TUDatasetInterface(TUDataset):
    r"""
    Class that wraps the :class:`torch_geometric.datasets.TUDataset` class to provide aliases of some fields.
    It implements the interface ``DatasetInterface`` but does not extend directly to avoid clashes of ``__init__`` methods
    """
    def __init__(self, root, name, transform=None, pre_transform=None, pre_filter=None, **kwargs):
        self.name = name
        # Do not call DatasetInterface __init__ method in this case, because otherwise it will break
        super().__init__(root=root, name=name,
                         transform=transform, pre_transform=pre_transform, pre_filter=pre_filter,
                         **kwargs)

    @property
    def dim_node_features(self):
        return self.num_node_features

    @property
    def dim_edge_features(self):
        return self.num_edge_features

    @property
    def dim_target(self):
        return self.num_classes

    def download(self):
        super().download()

    def process(self):
        super().process()

    def __len__(self) -> int:
        return len(self.data)


class PlanetoidDatasetInterface(Planetoid):
    r"""
    Class that wraps the :class:`torch_geometric.datasets.Planetoid` class to provide aliases of some fields.
    It implements the interface ``DatasetInterface`` but does not extend directly to avoid clashes of ``__init__`` methods
    """
    def __init__(self, root, name, transform=None, pre_transform=None, pre_filter=None, **kwargs):
        self.name = name
        # Do not call DatasetInterface __init__ method in this case, because otherwise it will break
        super().__init__(root=root, name=name,
                         transform=transform, pre_transform=pre_transform,
                         **kwargs)
    @property
    def dim_node_features(self):
        return self.num_node_features

    @property
    def dim_edge_features(self):
        return self.num_edge_features

    @property
    def dim_target(self):
        return self.num_classes

    def download(self):
        super().download()

    def process(self):
        super().process()

    def __len__(self) -> int:
        if isinstance(self.data, Data):
            return 1
        return len(self.data)


class OGBGDatasetInterface(PygGraphPropPredDataset):
    r"""
    Class that wraps the :class:`ogb.graphproppred.PygGraphPropPredDataset` class to provide aliases of some fields.
    It implements the interface ``DatasetInterface`` but does not extend directly to avoid clashes of ``__init__`` methods
    """
    def __init__(self, root, name, transform=None,
                 pre_transform=None, pre_filter=None, meta_dict=None):
        super().__init__('-'.join(name.split('_')), root, transform, pre_transform, meta_dict)  #
        self.name = name
        self.data.y = self.data.y.squeeze()

    def get_idx_split(self, split_type: str = None) -> dict:
        return self.dataset.get_idx_split(split_type=split_type)

    @property
    def dim_node_features(self):
        return 1

    @property
    def dim_edge_features(self):
        if self.data.edge_attr is not None:
            return self.data.edge_attr.shape[1]
        else:
            return 0

    @property
    def dim_target(self):
        return 37

    @property
    def processed_file_names(self):
        return ['data.pt']

    def download(self):
        url = self.meta_info['url']
        if decide_download(url):
            path = download_url(url, self.original_root)
            extract_zip(path, self.original_root)
            print(f'Removing {path}')
            os.unlink(path)
            print(f'Removing {self.root}')
            shutil.rmtree(self.root)
            print(f'Moving {os.path.join(self.original_root, self.download_name)} to {self.root}')
            shutil.move(os.path.join(self.original_root, self.download_name), self.root)

    def process(self):
        super().process()

    def __len__(self) -> int:
        return len(self.data)
